# -*- coding: utf-8 -*-

from twisted.internet import reactor
from twisted.trial.unittest import TestCase

from il2ds_middleware.protocol import ConsoleFactory as ClientFactory

from il2ds_middleware.ds_emulator.service import RootService as DSService
from il2ds_middleware.ds_emulator.protocol import ConsoleFactory as DSFactory


# class TestClientFactory(TestCase):

#     def setUp(self):
#         self.dsfactory = DSFactory()
#         self.dsservice = DSService(self.dsfactory)
#         self.dsport = reactor.listenTCP(0, self.dsfactory)

#         self.addCleanup(self.dsport.stopListening)
#         self.addCleanup(self.dsservice.stopService)

#         self.dsservice.startService()

#     def test_connect_with_wrong_address(self):
#         pass



# -*- coding: utf-8 -*-

from zope.interface import implementer

from twisted.python import log

from il2ds_middleware.interfaces import ILineParser


@implementer(ILineParser)
class PropagatingLineParserMixing:

    propagate = False

    def _autopropagate(self, value=True):
        return False if self.propagate else value



# -*- coding: utf-8 -*-

import datetime

from twisted.application.internet import TimerService
from twisted.internet import defer


class LogWatchingService(TimerService):

    receiver = None
    log_file = None

    def __init__(self, log_path, interval):
        self.log_path = log_path
        TimerService.__init__(self, interval, self.do_watch)

    def do_watch(self):
        self.log_file.seek(self.log_file.tell())
        if self.receiver is not None:
            for line in self.log_file.readlines():
                time, data = self._parse_line(line)
                if data is not None:
                    self.receiver(data)

    def _parse_line(self, line):
        idx = line.find(']')
        time_raw = line[1:idx]
        format = "%I:%M:%S %p"
        try:
            time = datetime.datetime.strptime(time_raw, format)
        except:
            format = "%b %d, %Y " + format
            try:
                time = datetime.datetime.strptime(time_raw, format)
            except:
                return (None, None)
        finally:
            data = line[idx + 1:].lstrip()
            return time, data

    def startService(self):
        if self.log_path is None:
            return
        self.log_file = open(self.log_path, 'r')
        TimerService.startService(self)

    def stopService(self):
        if self.log_file is None:
            return defer.succeed(None)
        else:
            self.log_file.close()
            self.log_file = None
            return TimerService.stopService(self)



# -*- coding: utf-8 -*-

from zope.interface import Interface


class ILineParser(Interface):

    def parse_line(self, line):
        """Parse line.

        :param line: line to parse.
        :type line: str.

        :returns: bool -- True if line was successfully parsed, otherwise False.
        """



# -*- coding: utf-8 -*-

from twisted.internet.protocol import ServerFactory
from twisted.protocols.basic import LineReceiver
from twisted.python import log

from zope.interface import implementer

from il2ds_middleware.protocol import DeviceLinkProtocol
from il2ds_middleware.ds_emulator.interfaces import ILineBroadcaster

class DeviceLinkServerProtocol(DeviceLinkProtocol):

    on_requests = None

    def requests_received(self, requests, address):
        if self.on_requests is not None:
            self.on_requests(requests, address, self)


class ConsoleProtocol(LineReceiver):

    def connectionMade(self):
        self.factory.client_joined(self)

    def connectionLost(self, reason):
        self.factory.client_left(self)

    def lineReceived(self, line):
        self.factory.got_line(line)

    def message(self, message):
        self.sendLine(message + '\\n')


@implementer(ILineBroadcaster)
class ConsoleFactory(ServerFactory):

    protocol = ConsoleProtocol

    def __init__(self):
        self.clients = []

    def client_joined(self, client):
        self.clients.append(client)

    def client_left(self, client):
        self.clients.remove(client)

    def got_line(self, line):
        if self.receiver is not None:
            self.receiver(line)

    def broadcast_line(self, line):

        def do_broadcast(line):
            for client in self.clients:
                client.message(line)

        from twisted.internet import reactor
        reactor.callLater(0, do_broadcast, line)



# -*- coding: utf-8 -*-

from twisted.internet.defer import Deferred
from twisted.internet.protocol import ClientFactory
from twisted.protocols.basic import LineReceiver

from il2ds_middleware.protocol import DeviceLinkProtocol
from il2ds_middleware.ds_emulator.protocol import ConsoleFactory


class ConsoleServerFactory(ConsoleFactory):

    def __init__(self):
        ConsoleFactory.__init__(self)
        self.on_connection_lost = Deferred()

    def client_left(self, client):
        ConsoleFactory.client_left(self, client)
        self.on_connection_lost.callback(client)


class ConsoleClientProtocol(LineReceiver):

    def connectionMade(self):
        self.factory.client_joined(self)

    def connectionLost(self, reason):
        self.factory.client_left(self)

    def lineReceived(self, line):
        self.factory.got_line(line)


class ConsoleClientFactory(ClientFactory):

    protocol = ConsoleClientProtocol
    receiver = None

    def __init__(self):
        self.clients = []
        self.on_connection_made = Deferred()
        self.on_connection_lost = Deferred()

    def client_joined(self, client):
        self.on_connection_made.callback(client)
        self.clients.append(client)

    def client_left(self, client):
        self.on_connection_lost.callback(client)
        self.clients.remove(client)

    def got_line(self, line):
        if self.receiver is not None:
            self.receiver(line)

    def message(self, message):

        def do_message(message):
            for client in self.clients:
                client.sendLine(message)

        from twisted.internet import reactor
        reactor.callLater(0, do_message, message)


class DeviceLinkClientProtocol(DeviceLinkProtocol):

    receiver = None

    def answers_received(self, answers, address):
        if self.receiver is not None:
            line = "A/" + self._format(answers)
            self.receiver(line)



# -*- coding: utf-8 -*-

from twisted.internet import reactor
from twisted.internet.protocol import ClientFactory, DatagramProtocol
from twisted.protocols.basic import LineOnlyReceiver
from twisted.python import log

from il2ds_middleware.constants import (DEVICE_LINK_OPCODE,
    DEVICE_LINK_PREFIXES, DEVICE_LINK_CMD_SEPARATOR as DL_CMD_SEP,
    DEVICE_LINK_ARGS_SEPARATOR as DL_ARGS_SEP, )


class ConsoleProtocol(LineOnlyReceiver):

    def connectionMade(self):
        log.msg("Connection established with {0}".format(
            self.transport.getaddress()))

    def connectionLost(self, reason):
        log.err("Connection with {0} lost: {1}".format(
            self.transport.getaddress(), reason))

    def lineReceived(self, line):
        if line == '' or line == "exit\\n":
            log.err("Game server is shut down")
            reactor.stop()
            return
        if line.startswith("<consoleN>"):
            return
        if line.endswith("\\n"):
            line = line[:-2]
        if line.startswith("\\u0020"):
            line = " " + line[6:]


class ConsoleFactory(ClientFactory):

    protocol = ConsoleProtocol
    receiver = None

    def clientConnectionFailed(self, connector, reason):
        log.err("Connection failed: %s" % reason)


class DeviceLinkProtocol(DatagramProtocol):

    def __init__(self, address=None):
        self.address = address

    def datagramReceived(self, data, (host, port)):
        if data.startswith(DEVICE_LINK_PREFIXES['answer']):
            processor = self.answers_received
        elif data.startswith(DEVICE_LINK_PREFIXES['request']):
            processor = self.requests_received
        else:
            log.err("Malformed data from {0}: \"{1}\"")
            return
        prepared_data = data[1:].strip(DL_CMD_SEP)
        payloads = self._parse(prepared_data)
        processor(payloads, (host, port))

    def _parse(self, data):
        results = []
        for chunk in data.split(DL_CMD_SEP):
            command = chunk.split(DL_ARGS_SEP)
            result = {}
            result['command'] = command[0]
            args = command[1:]
            if args:
                result['args'] = args
            results.append(result)
        return results

    def answers_received(self, answers, address):
        raise NotImplementedError

    def requests_received(self, requests, address):
        raise NotImplementedError

    def send_request(self, request, address=None):
        self.send_requests([request, ], address)

    def send_requests(self, requests, address=None):
        self._send_payloads(DEVICE_LINK_PREFIXES['request'], requests, address)

    def send_answer(self, answer, address=None):
        self.send_answers([answer, ], address)

    def send_answers(self, answers, address=None):
        self._send_payloads(DEVICE_LINK_PREFIXES['answer'], answers, address)

    def _send_payloads(self, prefix, payloads, address=None):
        data = DL_CMD_SEP.join(
            [prefix, self._format(payloads), ])
        self.transport.write(data, address or self.address)

    def _format(self, payloads):
        chunks = []
        for payload in payloads:
            chunk = [payload['command'], ]
            raw_args = payload.get('args')
            if raw_args:
                chunk.extend([str(_) for _ in raw_args])
            chunks.append(DL_ARGS_SEP.join(chunk))
        return DL_CMD_SEP.join(chunks)



# -*- coding: utf-8 -*-

from twisted.internet import defer
from twisted.trial.unittest import TestCase

from il2ds_middleware.ds_emulator.service import RootService
from il2ds_middleware.ds_emulator.protocol import DeviceLinkServerProtocol
from il2ds_middleware.ds_emulator.tests.protocol import (ConsoleServerFactory,
    ConsoleClientFactory, ConsoleClientProtocol, DeviceLinkClientProtocol, )
from il2ds_middleware.ds_emulator.tests.service import LogWatchingService


class BaseTestCase(TestCase):

    log_watcher_interval = 0.1
    timeout_value = 0.05
    log_path = None

    def setUp(self):
        self.log_watcher = LogWatchingService(
            self.log_path, self.log_watcher_interval)
        self._listen_server()
        return self._connect_client()

    def _listen_server(self):
        self.console_server_factory = ConsoleServerFactory()
        self.service = RootService(self.console_server_factory, self.log_path)
        self.dl_server = DeviceLinkServerProtocol()

        self.console_server_factory.receiver = self.service.parse_line
        self.dl_server.on_requests = self.service.getServiceNamed(
            'dl').got_requests

        from twisted.internet import reactor
        self.console_server_port = reactor.listenTCP(
            0, self.console_server_factory, interface="127.0.0.1")
        self.dl_server_port = reactor.listenUDP(
            0, self.dl_server, interface="127.0.0.1")
        self.service.startService()

    def _connect_client(self):
        self.console_client_factory = ConsoleClientFactory()
        self.console_message = self.console_client_factory.message

        ds_host = self.dl_server_port.getHost()
        self.dl_client = DeviceLinkClientProtocol((ds_host.host, ds_host.port))

        from twisted.internet import reactor
        self.console_client_port = reactor.connectTCP(
            "127.0.0.1", self.console_server_port.getHost().port,
            self.console_client_factory)
        self.dl_client_port = reactor.listenUDP(0, self.dl_client)
        return self.console_client_factory.on_connection_made

    def tearDown(self):
        self.console_server_factory.on_data = None
        self.dl_server.on_request = None
        self.console_client_factory.receiver = None
        self.dl_client.receiver = None

        console_server_stopped = defer.maybeDeferred(
            self.console_server_port.stopListening)
        dl_server_stopped = defer.maybeDeferred(
            self.dl_server_port.stopListening)
        dl_client_stopped = defer.maybeDeferred(
            self.dl_client_port.stopListening)
        self.console_client_port.disconnect()

        return defer.gatherResults([
            console_server_stopped, dl_server_stopped, dl_client_stopped,
            self.console_client_factory.on_connection_lost,
            self.console_server_factory.on_connection_lost,
            self.log_watcher.stopService(), self.service.stopService(), ])

    def _make_timeout(self, callback):
        from twisted.internet import reactor
        return reactor.callLater(self.timeout_value, callback, None)

    def _get_unexpecting_line_receiver(self, d):

        def got_line(line):
            timeout.cancel()
            from twisted.trial.unittest import FailTest
            d.errback(FailTest("Unexpected data:\n\t{0}.".format(line)))

        timeout = self._make_timeout(d.callback)
        return got_line

    def _get_expecting_line_receiver(self, expected_lines, d):

        def got_line(line):
            try:
                self.assertEqual(line, expected_lines.pop(0))
            except Exception as e:
                timeout.cancel()
                d.errback(e)
            else:
                if expected_lines:
                    return
                timeout.cancel()
                d.callback(None)

        def on_timeout(_):
            from twisted.trial.unittest import FailTest
            d.errback(FailTest(
                'Timed out, remaining lines:\n\t'+'\n\t'.join(expected_lines)))

        timeout = self._make_timeout(on_timeout)
        return got_line

    def _set_console_expecting_receiver(self, expected_lines, d):
        self.console_client_factory.receiver = \
            self._get_expecting_line_receiver(expected_lines, d)

    def _set_dl_expecting_receiver(self, expected_lines, d):
        self.dl_client.receiver = \
            self._get_expecting_line_receiver(expected_lines, d)

    def _set_event_log_expecting_receiver(self, expected_lines, d):
        self.log_watcher.receiver = \
            self._get_expecting_line_receiver(expected_lines, d)
